from __future__ import annotations

from types import MappingProxyType
from typing import Any, Callable, Optional, Sequence, Union

from . import exceptions


class Empty:
    def __str__(self):
        return "EMPTY"


empty = Empty()


class Validator:
    default_messages = MappingProxyType(
        {
            "must_exist_true": "{name} is required in env {env}",
            "must_exist_false": "{name} cannot exists in env {env}",
            "condition": "{name} invalid for {function}({value}) in env {env}",
            "operations": (
                "{name} must {operation} {op_value} " "but it is {value} in env {env}"
            ),
            "combined": "combined validators failed {errors}",
        }
    )

    def __init__(
        self,
        *names: str,
        must_exist: Optional[bool] = None,
        required: Optional[bool] = None,  # alias for `must_exist`
        condition: Optional[Callable[[Any], bool]] = None,
        when: Optional[Validator] = None,
        env: Optional[Union[str, Sequence[str]]] = None,
        # messages: Optional[Dict[str, str]] = None,
        default: Optional[Union[Any, Callable[[Any, Validator], Any]]] = empty,
        description: Optional[str] = None,
        **operations: Any,
    ) -> None:
        self.messages = dict(self.default_messages)
        # if messages:
        #     self.messages.update(messages)

        self.names = names
        self.must_exist = must_exist if must_exist is not None else required
        self.condition = condition
        self.when = when
        self.operations = operations
        self.default = default
        self.description = description
        self.envs: Optional[Sequence[str]] = None

        if isinstance(env, str):
            self.envs = [env]
        elif isinstance(env, (list, tuple)):
            self.envs = env

    def validate(
        self,
        settings: Any,
        env: Optional[str] = None,
    ) -> None:
        print("self.must_exist", self.names, self.must_exist)
        env = env or settings.current_env
        for name in self.names:

            value = settings.get(name)

            if self.must_exist is True and value is None:
                raise exceptions.ValidationError(
                    self.messages["must_exist_true"].format(name=name, env=env)
                )


class ValidatorList(list):
    def __init__(
        self,
        settings: Any,
        validators: Optional[Sequence[Validator]] = None,
        *args: Validator,
        **kwargs: Any,
    ) -> None:
        if isinstance(validators, (list, tuple)):
            args = list(args) + list(validators)
        super(ValidatorList, self).__init__(args, **kwargs)  # type: ignore
        self.settings = settings

    def validate(
        self,
    ) -> None:
        for validator in self:
            validator.validate(self.settings)



from __future__ import annotations

from typing import Any, Dict


class Settings(dict):
    def __contains__(self, k: str) -> bool:
        try:
            return dict.__contains__(self, k) or hasattr(self, k)
        except Exception:
            return False

    def __getattr__(self, k: str) -> Any:
        try:
            # Throws exception if not in prototype chain
            return object.__getattribute__(self, k)
        except AttributeError:
            try:
                return self[k]
            except KeyError:
                raise AttributeError(k)

    def __setattr__(self, k: str, v: Any) -> None:
        try:
            # Throws exception if not in prototype chain
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                self[k] = v
            except Exception:
                raise AttributeError(k)
        else:
            object.__setattr__(self, k, v)

    def __delattr__(self, k: str) -> None:
        try:
            # Throws exception if not in prototype chain
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                del self[k]
            except KeyError:
                raise AttributeError(k)
        else:
            object.__delattr__(self, k)

    def as_dict(self) -> Dict[str, Any]:
        return settings_to_dict(self)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> Settings:
        return dict_to_settings(d)


def dict_to_settings(x: Dict[str, Any]) -> Settings:
    if isinstance(x, dict):
        return Settings((k, dict_to_settings(v)) for k, v in dict.items(x))
    elif isinstance(x, (list, tuple)):
        return type(x)(dict_to_settings(v) for v in x)
    else:
        return x


def settings_to_dict(x: Settings) -> Dict[str, Any]:
    if isinstance(x, dict):
        return dict((k, settings_to_dict(v)) for k, v in dict.items(x))
    elif isinstance(x, (list, tuple)):
        return type(x)(settings_to_dict(v) for v in x)
    else:
        return x



import os
from typing import List, Optional, Union

from confoid.exceptions import MixedFileTypes

from .base import BaseLoader
from .yaml_loader import YAMLLoader


def settings_loader(
    filenames=Union[str, List[str]], env_default_value=None
) -> Optional[BaseLoader]:
    exts = list(set([os.path.splitext(filename)[-1] for filename in filenames]))
    if not len(exts) == 1:  # all discovered files should have same extension
        raise MixedFileTypes("Settings files cannot contained mixed filetypes")

    exts = exts[0]
    if exts in YAMLLoader.EXTENSIONS:
        return YAMLLoader.loader(filenames, env_default_value=env_default_value)
    return None



from __future__ import annotations

import os
from typing import Dict, List, Optional, Union

from . import exceptions, loaders
from .base import Settings
from .constants import ENV_VARS
from .validator import Validator, ValidatorList

CONFIGS: Dict[str, Settings] = {}


class Resolver:
    def __init__(self, prefix: str, **kwargs) -> None:
        self.prefix = prefix
        self.kwargs = kwargs
        self.base_dir = ""
        self.current_env = "local"
        self.files: Optional[Union[str, List[str]]] = None
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.resolve()
        self.expand()

    @staticmethod
    def to_full_path(base_dir: str, path: str) -> str:
        return os.path.join(base_dir, path)

    def setattr_from_env(
        self, key, env: str, default: Optional[str] = ""
    ) -> Union[str, None]:
        _value = os.getenv(f"{self.prefix}_{env}")
        if _value:
            setattr(self, key, _value)
        if not getattr(
            self, key, None
        ):  # value is still None then set fallback (default) value
            setattr(self, key, default)
        return os.getenv(f"{self.prefix}_{env}", default)

    def resolve(self) -> Resolver:
        for key, value in ENV_VARS.items():
            self.setattr_from_env(key, value[0], default=value[1])
        return self

    def expand(self) -> Resolver:
        if not self.files:
            return self

        files = self.files
        if isinstance(files, str):
            files = [files]
        files = [self.to_full_path(self.base_dir, x) for x in files]
        valid_files = []
        for _file in files:
            if not os.path.exists(_file):
                print(f"{_file}: does not exists.. skipping")
            else:
                valid_files.append(_file)
        if not valid_files:
            error = f"No valid files found, checked {files}"
            raise exceptions.ResolveError(error)

        if len(valid_files) == 1 and self.current_env:
            name, ext = os.path.splitext(valid_files[0])
            env_file = f"{name}.{self.current_env}.{ext[1:]}"
            if os.path.exists(env_file):
                valid_files.append(env_file)
        self.files = valid_files
        return self

    def validate(self) -> Resolver:
        for key, _ in self.kwargs.items():
            _value = getattr(self, key)
            if _value is None:
                env_var = f"{self.prefix}_{ENV_VARS[key][0]}"
                error = f"No {key} provided to config also checked env: {env_var}"
                raise exceptions.ResolveError(error)
        return self


def Config(
    prefix: str,
    files: Optional[Union[str, List[str]]] = None,
    base_dir: Optional[str] = "",
    current_env: Optional[str] = "local",
    validators: Optional[List[Validator]] = [],
    reload: bool = False,
    env_default_value: Optional[str] = "N/A",
) -> Union[bool, Settings]:  # noqa

    if not reload and prefix in CONFIGS:
        return CONFIGS[prefix]

    resolver = Resolver(
        prefix, base_dir=base_dir, current_env=current_env, files=files
    ).validate()

    print(f"{prefix}: Setting environment: {resolver.current_env}")
    print(f"{prefix}: Using config files: {resolver.files}")

    data = loaders.settings_loader(resolver.files, env_default_value=env_default_value)
    settings = Settings.from_dict(data)
    settings.current_env = resolver.current_env
    ValidatorList(settings, validators).validate()
    CONFIGS[prefix] = settings

    return settings



import os
import re
from typing import List, Optional

import yaml
from confoid import constants
from confoid.loaders.base import BaseLoader

default_sep = ":"
default_value = "N/A"
raise_if_na = False

default_sep = default_sep or ""
default_sep_pattern = r"(" + default_sep + "[^}]+)?" if default_sep else ""
pattern = re.compile(
    r".*?\$\{([^}{" + default_sep + r"]+)" + default_sep_pattern + r"\}.*?"
)


class ENV(dict):
    """Env class."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        for key, value in kwargs.items():
            if not os.getenv(key):
                os.environ[key] = value

    def to_dict(self):
        return self.kwargs


class YAMLLoader(BaseLoader):
    EXTENSIONS = constants.YAML_EXTENSIONS

    def __init__(self, env_default_value=None) -> None:
        self.loader = yaml.SafeLoader
        self.encoding = "utf-8"
        self.tag = "!ENV"
        self.loader.add_constructor("!Environment", self.env_constructor)
        self.loader.add_implicit_resolver(self.tag, pattern, None)
        self.loader.add_constructor(self.tag, self.constructor_env_variables)

        if env_default_value is not None:
            global default_value
            default_value = env_default_value

    @classmethod
    def loader(cls, filenames: List[str], env_default_value=None):
        data = {}
        loader = cls(env_default_value=env_default_value)
        for filename in filenames:
            _data = loader.load(filename)
            data = loader.update(data, _data)
        return data

    def load(self, filename: str, encoding: Optional[str] = None):
        settings = {}
        with open(filename, encoding=encoding or self.encoding) as conf_data:
            settings = yaml.load(conf_data, Loader=self.loader)  # nosec
        return settings

    @staticmethod
    def env_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> ENV:
        """Construct an env node."""
        return ENV(**loader.construct_mapping(node)).to_dict()

    @staticmethod
    def clean_value(value):
        if value.startswith(('"', "'")):
            value = value[1:]

        if value.endswith(('"', "'")):
            value = value[:-1]
        return value

    @staticmethod
    def get_env_fallback(g, key, default_value):
        curr_default_value = default_value
        found = False
        for each in g:
            if default_sep in each:
                _, curr_default_value = each.split(default_sep, 1)
                if curr_default_value:
                    curr_default_value = YAMLLoader.clean_value(curr_default_value)
                    found = True
                    break
        if not found and raise_if_na:
            raise ValueError(f"Could not find default value for {key}")
        return curr_default_value

    @staticmethod
    def constructor_env_variables(loader, node):
        value = loader.construct_scalar(node)
        match = pattern.findall(value)  # to find all env variables in line
        if not match:
            return value

        full_value = value
        for g in match:
            curr_default_value = default_value
            env_var_name = g
            env_var_name_with_default = g
            if default_sep and isinstance(g, tuple) and len(g) > 1:
                env_var_name = g[0]
                env_var_name_with_default = "".join(g)
                curr_default_value = YAMLLoader.get_env_fallback(
                    g, env_var_name, curr_default_value
                )
            full_value = full_value.replace(
                f"${{{env_var_name_with_default}}}",
                os.environ.get(env_var_name, curr_default_value),
            )
        return full_value



YAML_EXTENSIONS = (".yaml", ".yml")

ALL_EXTENSIONS = YAML_EXTENSIONS

ENV_VARS = {
    "base_dir": ["CONFIG_BASEDIR", ""],
    "current_env": ["ENV", "local"],
    "files": ["CONFIG_FULLPATH", None],
}



"""Package setup."""

from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="confoid",
    description="confoid",
    version="0.1.0",
    author="Savio Fernandes",
    author_email="savio@saviof.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/saviof-hub/confoid",
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=["pyyaml==6.0"],
    extras_require={
        "dev": [
            'coverage[toml]',
            'pytest',
            'pytest-asyncio',
            'pytest-cov',
            'flake8',
            'flake8-docstrings',
            'black',
            'isort',
            'requests',
            'hypothesis',
            'mypy',
            'bandit',
            'pylint',
            "twine"
        ]
    },
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import collections.abc
from typing import Optional


class BaseLoader(object):
    EXTENSIONS = ()

    def load(self, filename: str, encoding: Optional[str] = None) -> None:
        raise NotImplementedError

    def get_key(self, d, k, v):
        val = d.get(k, {})
        if val is None and v:
            return {}
        return val

    def update(self, d, u):
        if not u:
            return d
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = self.update(self.get_key(d, k, v), v)
            else:
                d[k] = v
        return d
